from ._sync.core import Swarm
from ._sync.types import Agent, Response
from ._async.core import AsyncSwarm
from ._async.types import AsyncAgent, AsyncResponse

__all__ = ["Swarm", "Agent", "Response", "AsyncSwarm", "AsyncAgent", "AsyncResponse"]



# Code generated from codegen/templates/swarm/(sync)/types.py.jinja2. DO NOT EDIT MANUALLY!


from pydantic import BaseModel
from openai.types.chat import (
    ChatCompletionMessage,
    ChatCompletionChunk,
    ChatCompletion,
)
from openai import Stream
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)
from typing import (
    Any,
    Generator,
    List,
    Callable,
    TypedDict,
    Union,
    Optional,
    TypeAlias,
    Literal,
    Coroutine,
    Dict,
)

Message : TypeAlias = Dict[Literal[
    "content",
    "sender",
    "role",
    "function_call",
    "tool_calls",
    "tool_call_id",
    "tool_name",
    "audio",
    "refusal",
], Any]


class RetryState(BaseModel):
  tries: int = 0
  sleep_time: int = 0
  error: str = ""

AgentFunction = Callable[..., Union[str, "Agent", dict, "Result"]]
AgentInstructions: TypeAlias = Union[str, Callable[..., str]]

class Agent(BaseModel):
    name: str = "Agent"
    model: str = "gpt-4o"
    instructions: AgentInstructions = "You are a helpful agent."  # type: ignore
    functions: List[AgentFunction] = []
    tool_choice: Optional[str] = None
    parallel_tool_calls: bool = True

class Response(BaseModel):
    messages: List[Message] = []
    agent: Optional[Agent] = None
    context_variables: dict = {}


class Result(BaseModel):
    """
    Encapsulates the possible return values for an agent function.

    Attributes:
        value (str): The result value as a string.
        agent (Agent): The agent instance, if applicable.
        context_variables (dict): A dictionary of context variables.
    """

    value: str = ""
    agent: Optional[Agent] = None
    context_variables: dict = {}


DelimStreamingChunk = dict[Literal["delim"], Literal["start", "end"]]
ResponseStreamingChunk = dict[Literal["response"], Response]

class DeltaResponseStreamingChunk(TypedDict):
    partial_response: Response

class MessageStreamingChunk(TypedDict):
    content: Any
    sender: str
    role: Literal["system", "user", "assistant", "tool"]
    tool_calls: List[dict[Literal["id", "function", "type"], Any]]
    function_call: None


StreamingResponse: TypeAlias = Generator[
    Union[
        DelimStreamingChunk,
        MessageStreamingChunk,
        ResponseStreamingChunk,
        DeltaResponseStreamingChunk,
    ],
    None,
    None,
]


import datetime
import random

import database
from swarm import Agent
from swarm.agents import create_triage_agent
from swarm.repl import run_demo_loop


def refund_item(user_id, item_id):
    """Initiate a refund based on the user ID and item ID.
    Takes as input arguments in the format '{"user_id":"1","item_id":"3"}'
    """
    conn = database.get_connection()
    cursor = conn.cursor()
    cursor.execute(
        """
        SELECT amount FROM PurchaseHistory
        WHERE user_id = ? AND item_id = ?
    """,
        (user_id, item_id),
    )
    result = cursor.fetchone()
    if result:
        amount = result[0]
        print(f"Refunding ${amount} to user ID {user_id} for item ID {item_id}.")
    else:
        print(f"No purchase found for user ID {user_id} and item ID {item_id}.")
    print("Refund initiated")


def notify_customer(user_id, method):
    """Notify a customer by their preferred method of either phone or email.
    Takes as input arguments in the format '{"user_id":"1","method":"email"}'"""

    conn = database.get_connection()
    cursor = conn.cursor()
    cursor.execute(
        """
        SELECT email, phone FROM Users
        WHERE user_id = ?
    """,
        (user_id,),
    )
    user = cursor.fetchone()
    if user:
        email, phone = user
        if method == "email" and email:
            print(f"Emailed customer {email} a notification.")
        elif method == "phone" and phone:
            print(f"Texted customer {phone} a notification.")
        else:
            print(f"No {method} contact available for user ID {user_id}.")
    else:
        print(f"User ID {user_id} not found.")


def order_item(user_id, product_id):
    """Place an order for a product based on the user ID and product ID.
    Takes as input arguments in the format '{"user_id":"1","product_id":"2"}'"""
    date_of_purchase = datetime.datetime.now()
    item_id = random.randint(1, 300)

    conn = database.get_connection()
    cursor = conn.cursor()
    cursor.execute(
        """
        SELECT product_id, product_name, price FROM Products
        WHERE product_id = ?
    """,
        (product_id,),
    )
    result = cursor.fetchone()
    if result:
        product_id, product_name, price = result
        print(
            f"Ordering product {product_name} for user ID {user_id}. The price is {price}."
        )
        # Add the purchase to the database
        database.add_purchase(user_id, date_of_purchase, item_id, price)
    else:
        print(f"Product {product_id} not found.")


# Initialize the database
database.initialize_database()

# Preview tables
database.preview_table("Users")
database.preview_table("PurchaseHistory")
database.preview_table("Products")

# Define the agents

refunds_agent = Agent(
    name="Refunds Agent",
    description=f"""You are a refund agent that handles all actions related to refunds after a return has been processed.
    You must ask for both the user ID and item ID to initiate a refund. Ask for both user_id and item_id in one message.
    If the user asks you to notify them, you must ask them what their preferred method of notification is. For notifications, you must
    ask them for user_id and method in one message.""",
    functions=[refund_item, notify_customer],
)

sales_agent = Agent(
    name="Sales Agent",
    description=f"""You are a sales agent that handles all actions related to placing an order to purchase an item.
    Regardless of what the user wants to purchase, must ask for BOTH the user ID and product ID to place an order.
    An order cannot be placed without these two pieces of information. Ask for both user_id and product_id in one message.
    If the user asks you to notify them, you must ask them what their preferred method is. For notifications, you must
    ask them for user_id and method in one message.
    """,
    functions=[order_item, notify_customer],
)

triage_agent = create_triage_agent(
    name="Triage Agent",
    instructions=f"""You are to triage a users request, and call a tool to transfer to the right intent.
    Once you are ready to transfer to the right intent, call the tool to transfer to the right intent.
    You dont need to know specifics, just the topic of the request.
    If the user request is about making an order or purchasing an item, transfer to the Sales Agent.
    If the user request is about getting a refund on an item or returning a product, transfer to the Refunds Agent.
    When you need more information to triage the request to an agent, ask a direct question without explaining why you're asking it.
    Do not share your thought process with the user! Do not make unreasonable assumptions on behalf of user.""",
    agents=[sales_agent, refunds_agent],
    add_backlinks=True,
)

for f in triage_agent.functions:
    print(f.__name__)

if __name__ == "__main__":
    # Run the demo loop
    run_demo_loop(triage_agent, debug=False)



import inspect
from datetime import datetime
from typing import Any, List, get_origin, get_args, Callable, Dict
from pydantic import BaseModel, TypeAdapter


def debug_print(debug: bool, *args: Any) -> None:
    if not debug:
        return
    timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    message = " ".join(map(str, args))
    print(f"\033[97m[\033[90m{timestamp}\033[97m]\033[90m {message}\033[0m")


def merge_fields(target, source):
    for key, value in source.items():
        if isinstance(value, str):
            target[key] += value
        elif value is not None and isinstance(value, dict):
            merge_fields(target[key], value)


def merge_chunk(final_response: dict, delta: dict) -> None:
    delta.pop("role", None)
    merge_fields(final_response, delta)

    tool_calls = delta.get("tool_calls")
    if tool_calls and len(tool_calls) > 0:
        index = tool_calls[0].pop("index")
        if "type" in tool_calls[0]:
            final_response["tool_calls"][index]["type"] = tool_calls[0].pop("type")
        merge_fields(final_response["tool_calls"][index], tool_calls[0])


def remove_titles(schema: dict | list) -> dict | list:
    """
    Recursively removes 'title' keys from the schema dictionary.

    Args:
        schema: The schema dictionary from which to remove 'title' keys.

    Returns:
        The schema dictionary without 'title' keys.
    """
    if isinstance(schema, dict):
        schema.pop('title', None)
        for key, value in schema.items():
            schema[key] = remove_titles(value)
    elif isinstance(schema, list):
        schema = [remove_titles(item) for item in schema]
    return schema

def function_to_json(func) -> dict:
    """
    Converts a Python function into a JSON-serializable dictionary
    that describes the function's signature, including its name,
    description, and parameters.

    Args:
        func: The function to be converted.

    Returns:
        A dictionary representing the function's signature in JSON format.
    """
    type_map = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
        list: "array",
        dict: "object",
        type(None): "null",
    }


    try:
        signature = inspect.signature(func)
    except ValueError as e:
        raise ValueError(
            f"Failed to get signature for function {func.__name__}: {str(e)}"
        )

    parameters = {}
    for param in signature.parameters.values():
        annotation = param.annotation
        try:
            origin = get_origin(annotation)
            args = get_args(annotation)

            if isinstance(annotation, type) and issubclass(annotation, BaseModel):
                schema = annotation.model_json_schema()
                schema = remove_titles(schema)
                parameters[param.name] = schema
            elif origin is list and len(args) == 1:
                item_type = args[0]
                if isinstance(item_type, type) and issubclass(item_type, BaseModel):
                    # Handle list of BaseModel subclasses
                    item_schema = item_type.model_json_schema()
                    item_schema = remove_titles(item_schema)
                    parameters[param.name] = {
                        "type": "array",
                        "items": item_schema,
                    }
                elif item_type in type_map:
                    # Handle list of primitive types
                    parameters[param.name] = {
                        "type": "array",
                        "items": {"type": type_map.get(item_type, "string")},
                    }
                else:
                    # Default to string if type is unknown
                    parameters[param.name] = {
                        "type": "array",
                        "items": {"type": "string"},
                    }
            else:
                parameters[param.name] = {"type": type_map.get(annotation, "string")}
        except KeyError as e:
            raise KeyError(
                f"Unknown type annotation {annotation} for parameter {param.name}: {str(e)}"
            )

    required = [
        param.name
        for param in signature.parameters.values()
        if param.default == inspect._empty
    ]

    return {
        "type": "function",
        "function": {
            "name": func.__name__,
            "description": func.__doc__ or "",
            "parameters": {
                "type": "object",
                "properties": parameters,
                "required": required,
            },
        },
    }

def json_to_function_args(func: Callable, json_args: Dict[str, Any]) -> Dict[str, Any]:
    """Converts a JSON object to function arguments based on the function's annotations."""
    sig = inspect.signature(func)
    args = {}
    for name, param in sig.parameters.items():
        if name in json_args:
            annotation = param.annotation
            value = json_args[name]
            if inspect.isclass(annotation) and issubclass(annotation, BaseModel):
                args[name] = annotation(**value)
            elif (
                getattr(annotation, '__origin__', None) is list and
                hasattr(annotation, '__args__') and
                len(annotation.__args__) > 0
            ):
                item_type = annotation.__args__[0]
                if inspect.isclass(item_type) and issubclass(item_type, BaseModel):
                    args[name] = [item_type(**item) for item in value]
                elif item_type in [str, int, float, bool]:
                    args[name] = value  # Assuming the list is of the correct primitive type
                else:
                    args[name] = value  # Default to passing the value as is
            else:
                args[name] = value
        else:
            args[name] = param.default
    return args



import sqlite3

# global connection
conn = None


def get_connection():
    global conn
    if conn is None:
        conn = sqlite3.connect("application.db")
    return conn


def create_database():
    # Connect to a single SQLite database
    conn = get_connection()
    cursor = conn.cursor()

    # Create Users table
    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS Users (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            user_id INTEGER,
            first_name TEXT,
            last_name TEXT,
            email TEXT UNIQUE,
            phone TEXT
        )
    """
    )

    # Create PurchaseHistory table
    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS PurchaseHistory (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            user_id INTEGER,
            date_of_purchase TEXT,
            item_id INTEGER,
            amount REAL,
            FOREIGN KEY (user_id) REFERENCES Users(user_id)
        )
    """
    )

    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS Products (
            product_id INTEGER PRIMARY KEY,
            product_name TEXT NOT NULL,
            price REAL NOT NULL
        );
        """
    )

    # Save (commit) the changes
    conn.commit()


def add_user(user_id, first_name, last_name, email, phone):
    conn = get_connection()
    cursor = conn.cursor()

    # Check if the user already exists
    cursor.execute("SELECT * FROM Users WHERE user_id = ?", (user_id,))
    if cursor.fetchone():
        return

    try:
        cursor.execute(
            """
            INSERT INTO Users (user_id, first_name, last_name, email, phone)
            VALUES (?, ?, ?, ?, ?)
        """,
            (user_id, first_name, last_name, email, phone),
        )

        conn.commit()
    except sqlite3.Error as e:
        print(f"Database Error: {e}")


def add_purchase(user_id, date_of_purchase, item_id, amount):
    conn = get_connection()
    cursor = conn.cursor()

    # Check if the purchase already exists
    cursor.execute(
        """
        SELECT * FROM PurchaseHistory
        WHERE user_id = ? AND item_id = ? AND date_of_purchase = ?
    """,
        (user_id, item_id, date_of_purchase),
    )
    if cursor.fetchone():
        # print(f"Purchase already exists for user_id {user_id} on {date_of_purchase} for item_id {item_id}.")
        return

    try:
        cursor.execute(
            """
            INSERT INTO PurchaseHistory (user_id, date_of_purchase, item_id, amount)
            VALUES (?, ?, ?, ?)
        """,
            (user_id, date_of_purchase, item_id, amount),
        )

        conn.commit()
    except sqlite3.Error as e:
        print(f"Database Error: {e}")


def add_product(product_id, product_name, price):
    conn = get_connection()
    cursor = conn.cursor()

    try:
        cursor.execute(
            """
        INSERT INTO Products (product_id, product_name, price)
        VALUES (?, ?, ?);
        """,
            (product_id, product_name, price),
        )

        conn.commit()
    except sqlite3.Error as e:
        print(f"Database Error: {e}")


def close_connection():
    global conn
    if conn:
        conn.close()
        conn = None


def preview_table(table_name):
    conn = sqlite3.connect("application.db")  # Replace with your database name
    cursor = conn.cursor()

    cursor.execute(f"SELECT * FROM {table_name} LIMIT 5;")  # Limit to first 5 rows

    rows = cursor.fetchall()

    for row in rows:
        print(row)

    conn.close()


# Initialize and load database
def initialize_database():
    global conn

    # Initialize the database tables
    create_database()

    # Add some initial users
    initial_users = [
        (1, "Alice", "Smith", "alice@test.com", "123-456-7890"),
        (2, "Bob", "Johnson", "bob@test.com", "234-567-8901"),
        (3, "Sarah", "Brown", "sarah@test.com", "555-567-8901"),
        # Add more initial users here
    ]

    for user in initial_users:
        add_user(*user)

    # Add some initial purchases
    initial_purchases = [
        (1, "2024-01-01", 101, 99.99),
        (2, "2023-12-25", 100, 39.99),
        (3, "2023-11-14", 307, 49.99),
    ]

    for purchase in initial_purchases:
        add_purchase(*purchase)

    initial_products = [
        (7, "Hat", 19.99),
        (8, "Wool socks", 29.99),
        (9, "Shoes", 39.99),
    ]

    for product in initial_products:
        add_product(*product)



from configs.agents import *
from swarm.repl import run_demo_loop

context_variables = {
    "customer_context": """Here is what you know about the customer's details:
1. CUSTOMER_ID: customer_12345
2. NAME: John Doe
3. PHONE_NUMBER: (123) 456-7890
4. EMAIL: johndoe@example.com
5. STATUS: Premium
6. ACCOUNT_STATUS: Active
7. BALANCE: $0.00
8. LOCATION: 1234 Main St, San Francisco, CA 94123, USA
""",
    "flight_context": """The customer has an upcoming flight from LGA (Laguardia) in NYC to LAX in Los Angeles.
The flight # is 1919. The flight departure date is 3pm ET, 5/21/2024.""",
}
if __name__ == "__main__":
    run_demo_loop(triage_agent, context_variables=context_variables, debug=True)



import pytest
from swarm import Swarm, Agent
from swarm.types import Response
from tests.mock_client import MockOpenAIClient, create_mock_response
from unittest.mock import Mock
import json
from pydantic import BaseModel
import inspect
from typing import Any, Callable, Dict, List

DEFAULT_RESPONSE_CONTENT = "sample response content"


@pytest.fixture
def mock_openai_client():
    m = MockOpenAIClient()
    m.set_response(
        create_mock_response({"role": "assistant", "content": DEFAULT_RESPONSE_CONTENT})
    )
    return m


def test_run_with_simple_message(mock_openai_client: MockOpenAIClient):
    agent = Agent()
    # set up client and run
    client = Swarm(client=mock_openai_client)
    messages = [{"role": "user", "content": "Hello, how are you?"}]
    response = client.run(agent=agent, messages=messages)
    assert isinstance(response, Response)

    # assert response content
    assert response.messages[-1]["role"] == "assistant"
    assert response.messages[-1]["content"] == DEFAULT_RESPONSE_CONTENT


def test_tool_call(mock_openai_client: MockOpenAIClient):
    expected_location = "San Francisco"

    # set up mock to record function calls
    get_weather_mock = Mock()

    def get_weather(location):
        get_weather_mock(location=location)
        return "It's sunny today."

    agent = Agent(name="Test Agent", functions=[get_weather])
    messages = [
        {"role": "user", "content": "What's the weather like in San Francisco?"}
    ]

    # set mock to return a response that triggers function call
    mock_openai_client.set_sequential_responses(
        [
            create_mock_response(
                message={"role": "assistant", "content": ""},
                function_calls=[
                    {"name": "get_weather", "args": {"location": expected_location}}
                ],
            ),
            create_mock_response(
                {"role": "assistant", "content": DEFAULT_RESPONSE_CONTENT}
            ),
        ]
    )

    # set up client and run
    client = Swarm(client=mock_openai_client)
    response = client.run(agent=agent, messages=messages)

    get_weather_mock.assert_called_once_with(location=expected_location)
    assert isinstance(response, Response)
    assert response.messages[-1]["role"] == "assistant"
    assert response.messages[-1]["content"] == DEFAULT_RESPONSE_CONTENT


def test_execute_tools_false(mock_openai_client: MockOpenAIClient):
    expected_location = "San Francisco"

    # set up mock to record function calls
    get_weather_mock = Mock()

    def get_weather(location):
        get_weather_mock(location=location)
        return "It's sunny today."

    agent = Agent(name="Test Agent", functions=[get_weather])
    messages = [
        {"role": "user", "content": "What's the weather like in San Francisco?"}
    ]

    # set mock to return a response that triggers function call
    mock_openai_client.set_sequential_responses(
        [
            create_mock_response(
                message={"role": "assistant", "content": ""},
                function_calls=[
                    {"name": "get_weather", "args": {"location": expected_location}}
                ],
            ),
            create_mock_response(
                {"role": "assistant", "content": DEFAULT_RESPONSE_CONTENT}
            ),
        ]
    )

    # set up client and run
    client = Swarm(client=mock_openai_client)
    response = client.run(agent=agent, messages=messages, execute_tools=False)
    assert isinstance(response, Response)
    print(response)

    # assert function not called
    get_weather_mock.assert_not_called()

    # assert tool call is present in last response
    tool_calls = response.messages[-1].get("tool_calls")
    assert tool_calls is not None and len(tool_calls) == 1
    tool_call = tool_calls[0]
    assert tool_call["function"]["name"] == "get_weather"
    assert json.loads(tool_call["function"]["arguments"]) == {
        "location": expected_location
    }


def test_handoff(mock_openai_client: MockOpenAIClient):
    def transfer_to_agent2():
        return agent2

    agent1 = Agent(name="Test Agent 1", functions=[transfer_to_agent2])
    agent2 = Agent(name="Test Agent 2")

    # set mock to return a response that triggers the handoff
    mock_openai_client.set_sequential_responses(
        [
            create_mock_response(
                message={"role": "assistant", "content": ""},
                function_calls=[{"name": "transfer_to_agent2"}],
            ),
            create_mock_response(
                {"role": "assistant", "content": DEFAULT_RESPONSE_CONTENT}
            ),
        ]
    )

    # set up client and run
    client = Swarm(client=mock_openai_client)
    messages = [{"role": "user", "content": "I want to talk to agent 2"}]
    response = client.run(agent=agent1, messages=messages)
    assert isinstance(response, Response)

    assert response.agent == agent2
    assert response.messages[-1]["role"] == "assistant"
    assert response.messages[-1]["content"] == DEFAULT_RESPONSE_CONTENT


# Define Pydantic models
class WeatherRequest(BaseModel):
    location: str


def test_tool_with_pydantic_arguments(mock_openai_client: MockOpenAIClient):
    expected_location = "New York"

    # Define the Pydantic-based function
    get_weather_mock = Mock()
    def get_weather(request: WeatherRequest):
        get_weather_mock(request=request)
        return f"The weather in {request.location} is sunny."

    agent = Agent(name="Weather Agent", functions=[get_weather])

    messages = [
        {"role": "user", "content": f"What's the weather like in {expected_location}?"}
    ]

    mock_openai_client.set_sequential_responses(
        [
            create_mock_response(
                message={"role": "assistant", "content": ""},
                function_calls=[
                    {"name": "get_weather", "args": {"request": {"location": expected_location}}}
                ],
            ),
            create_mock_response(
                {"role": "assistant", "content": DEFAULT_RESPONSE_CONTENT}
            ),
        ]
    )

    client = Swarm(client=mock_openai_client)
    response = client.run(agent=agent, messages=messages)

    called_args = get_weather_mock.call_args
    print("called_args", called_args)
    args, kwargs = called_args
    assert isinstance(kwargs['request'], WeatherRequest)
    assert kwargs['request'].location == expected_location



# Code generated from codegen/templates/swarm/(sync)/core.py.jinja2. DO NOT EDIT MANUALLY!

# Standard library imports
import copy
import json
from collections import defaultdict
import time
from typing import (
    Any,
    Callable,
    Generator,
    List,
    Optional,
    Union,
    cast,
)
import time

# Package/library imports
from openai import OpenAI


# Local imports
from swarm.util import function_to_json, debug_print, merge_chunk, json_to_function_args
from .types import (
    Agent,
    AgentFunction,
    ChatCompletionChunk,
    ChatCompletion,
    ChatCompletionMessageToolCall,
    Function,
    Response,
    Result,
    RetryState,
    Stream,
    StreamingResponse,
    MessageStreamingChunk,
    Message,
)

__CTX_VARS_NAME__ = "context_variables"
RETRY_TIMES = [0, 5, 10, 30, 60]

class Swarm:
    def __init__(
        self,
        client=None,
        exponential_backoff: bool = False,
        retry_callback: Optional[Callable[[RetryState], Any]] = None,
    ):
        if not client:
            client = OpenAI()
        self.client = client
        self.exponential_backoff = exponential_backoff
        self.retry_callback = retry_callback

    def get_chat_completion(
        self,
        agent: Agent,
        history: List,
        context_variables: dict,
        model_override: str | None,
        stream: bool,
        debug: bool,
    ) -> ChatCompletion | Stream[ChatCompletionChunk]:
        context_variables = defaultdict(str, context_variables)
        instructions = (
            agent.instructions(context_variables)
            if callable(agent.instructions)
            else agent.instructions
        )
        # history contains "sender", which is unsupported
        history = [
            {
                k: v for k, v in message.items()
                if k not in ["sender", "tool_name"]
            } for message in history
        ]
        messages = [{"role": "system", "content": instructions}] + history
        debug_print(debug, "Getting chat completion for...:", messages)

        tools = [function_to_json(f) for f in agent.functions]
        # hide context_variables from model
        for tool in tools:
            params = tool["function"]["parameters"]
            params["properties"].pop(__CTX_VARS_NAME__, None)
            if __CTX_VARS_NAME__ in params["required"]:
                params["required"].remove(__CTX_VARS_NAME__)

        create_params = {
            "model": model_override or agent.model,
            "messages": messages,
            "tools": tools or None,
            "tool_choice": agent.tool_choice,
            "stream": stream,
        }

        if tools:
            create_params["parallel_tool_calls"] = agent.parallel_tool_calls

        return self.client.chat.completions.create(**create_params)

    def handle_function_result(self, result: Union[str, Agent, dict, Result], debug: bool) -> Result:
        match result:
            case Result() as result:
                return result

            case Agent() as agent:
                return Result(
                    value=json.dumps({"assistant": agent.name}),
                    agent=agent,
                )
            case _:
                try:
                    return Result(value=str(result))
                except Exception as e:
                    error_message = f"Failed to cast response to string: {result}. Make sure agent functions return a string or Result object. Error: {str(e)}"
                    debug_print(debug, error_message)
                    raise TypeError(error_message)

    def handle_tool_calls(
        self,
        tool_calls: List[ChatCompletionMessageToolCall],
        functions: List[AgentFunction],
        context_variables: dict,
        debug: bool,
    ) -> Response:
        function_map = {f.__name__: f for f in functions}
        partial_response = Response(
            messages=[], agent=None, context_variables={})

        assert partial_response.messages is not None, "partial_response.messages is None"

        for tool_call in tool_calls:
            name = tool_call.function.name
            # handle missing tool case, skip to next tool
            if name not in function_map:
                debug_print(debug, f"Tool {name} not found in function map.")
                partial_response.messages.append(
                    {
                        "role": "tool",
                        "tool_call_id": tool_call.id,
                        "tool_name": name,
                        "content": f"Error: Tool {name} not found.",
                    }
                )
                continue
            args = json_to_function_args(function_map[name], json.loads(tool_call.function.arguments))
            debug_print(
                debug, f"Processing tool call: {name} with arguments {args}")

            func = function_map[name]
            # pass context_variables to agent functions
            if __CTX_VARS_NAME__ in func.__code__.co_varnames:
                args[__CTX_VARS_NAME__] = context_variables
            raw_result = function_map[name](**args)

            result: Result = self.handle_function_result(raw_result, debug)
            partial_response.messages.append(
                {
                    "role": "tool",
                    "tool_call_id": tool_call.id,
                    "tool_name": name,
                    "content": result.value,
                }
            )
            partial_response.context_variables.update(result.context_variables)
            if result.agent:
                partial_response.agent = result.agent

        return partial_response

    def _get_streaming_completion_with_retry(
        self,
        active_agent: Agent,
        history: List[Message],
        context_variables: dict,
        model_override: Optional[str],
        debug: bool,
        message: dict
    ) -> StreamingResponse:
        tries = 0
        while True:
            try:
                completion = self.get_chat_completion(
                    agent=active_agent,
                    history=history,
                    context_variables=context_variables,
                    model_override=model_override,
                    stream=True,
                    debug=debug,
                )
                assert hasattr(completion, '__iter__') and hasattr(completion, '__next__'), "Expected generator for streaming completion"
                completion = cast(Stream[ChatCompletionChunk], completion)
                yielded_start = False

                for chunk in completion:
                    if not yielded_start:
                        yield {"delim": "start"}
                        yielded_start = True

                    chunk = cast(ChatCompletionChunk, chunk)
                    delta = cast(MessageStreamingChunk, chunk.choices[0].delta.model_dump())
                    if delta["role"] == "assistant":
                        delta["sender"] = active_agent.name
                    yield delta
                    to_merge = cast(dict, delta.copy())
                    to_merge.pop("role", None)
                    to_merge.pop("sender", None)
                    merge_chunk(message, to_merge)

                yield {"delim": "end"}
                break
            except Exception as e:
                if self.retry_callback:
                    self.retry_callback(
                        RetryState(
                            error=str(e),
                            tries=tries,
                            sleep_time=RETRY_TIMES[tries] if self.exponential_backoff else 0,
                        )
                    )
                tries += 1
                if tries >= len(RETRY_TIMES) or not self.exponential_backoff:
                    raise e
                time.sleep(RETRY_TIMES[tries])

    def run_and_stream(
        self,
        agent: Agent,
        messages: List[Message],
        context_variables: dict = {},
        model_override: Optional[str] = None,
        debug: bool = False,
        max_turns: int | float = float("inf"),
        execute_tools: bool = True,
    ) -> StreamingResponse:
        active_agent = agent
        context_variables = copy.deepcopy(context_variables)
        history = copy.deepcopy(messages)
        init_len = len(messages)

        while len(history) - init_len < max_turns:

            message = {
                "content": "",
                "sender": agent.name,
                "role": "assistant",
                "function_call": None,
                "tool_calls": defaultdict(
                    lambda: {
                        "function": {"arguments": "", "name": ""},
                        "id": "",
                        "type": "",
                    }
                ),
            }

            for chunk in self._get_streaming_completion_with_retry(
                active_agent=active_agent,
                history=history,
                context_variables=context_variables,
                model_override=model_override,
                debug=debug,
                message=message
            ):
                yield chunk

            message["tool_calls"] = list(
                message.get("tool_calls", {}).values())
            if not message["tool_calls"]:
                message["tool_calls"] = None
            debug_print(debug, "Received completion:", message)
            message["sender"] = active_agent.name
            history.append(cast(Message, message))

            if not message["tool_calls"] or not execute_tools:
                debug_print(debug, "Ending turn.")
                break

            # convert tool_calls to objects
            tool_calls = []
            for tool_call in message["tool_calls"]:
                function = Function(
                    arguments=tool_call["function"]["arguments"],
                    name=tool_call["function"]["name"],
                )
                tool_call_object = ChatCompletionMessageToolCall(
                    id=tool_call["id"], function=function, type=tool_call["type"]
                )
                tool_calls.append(tool_call_object)

            # handle function calls, updating context_variables, and switching agents
            partial_response = self.handle_tool_calls(
                tool_calls, active_agent.functions, context_variables, debug
            )
            yield {"partial_response": partial_response}
            history.extend(partial_response.messages)
            context_variables.update(partial_response.context_variables)
            if partial_response.agent:
                active_agent = partial_response.agent

        yield {
            "response": Response(
                messages=history[init_len:],
                agent=active_agent,
                context_variables=context_variables,
            )
        }

    def run(
        self,
        agent: Agent,
        messages: List,
        context_variables: dict = {},
        model_override: Optional[str] = None,
        stream: bool = False,
        debug: bool = False,
        max_turns: int | float = float("inf"),
        execute_tools: bool = True,
    ) -> Response | StreamingResponse:
        if stream:
            return self.run_and_stream(
                agent=agent,
                messages=messages,
                context_variables=context_variables,
                model_override=model_override,
                debug=debug,
                max_turns=max_turns,
                execute_tools=execute_tools,
            )
        active_agent = agent
        context_variables = copy.deepcopy(context_variables)
        history = copy.deepcopy(messages)
        init_len = len(messages)

        while len(history) - init_len < max_turns and active_agent:

            # get completion with current history, agent
            completion = self.get_chat_completion(
                agent=active_agent,
                history=history,
                context_variables=context_variables,
                model_override=model_override,
                stream=stream,
                debug=debug,
            )
            completion = cast(ChatCompletion, completion)
            message = completion.choices[0].message
            debug_print(debug, "Received completion:", message)
            message_json = message.model_dump()
            message_json["sender"] = active_agent.name
            history.append(message_json)

            if not message.tool_calls or not execute_tools:
                debug_print(debug, "Ending turn.")
                break

            # handle function calls, updating context_variables, and switching agents
            partial_response = self.handle_tool_calls(
                message.tool_calls, active_agent.functions, context_variables, debug
            )
            history.extend(partial_response.messages)
            context_variables.update(partial_response.context_variables)
            if partial_response.agent:
                active_agent = partial_response.agent

        return Response(
            messages=history[init_len:],
            agent=active_agent,
            context_variables=context_variables,
        )



from unittest.mock import MagicMock
from swarm.types import ChatCompletionMessage, ChatCompletionMessageToolCall, Function
from openai import OpenAI
from openai.types.chat.chat_completion import ChatCompletion, Choice
import json


def create_mock_response(message, function_calls=[], model="gpt-4o"):
    role = message.get("role", "assistant")
    content = message.get("content", "")
    tool_calls = (
        [
            ChatCompletionMessageToolCall(
                id="mock_tc_id",
                type="function",
                function=Function(
                    name=call.get("name", ""),
                    arguments=json.dumps(call.get("args", {})),
                ),
            )
            for call in function_calls
        ]
        if function_calls
        else None
    )

    return ChatCompletion(
        id="mock_cc_id",
        created=1234567890,
        model=model,
        object="chat.completion",
        choices=[
            Choice(
                message=ChatCompletionMessage(
                    role=role, content=content, tool_calls=tool_calls
                ),
                finish_reason="stop",
                index=0,
            )
        ],
    )


class MockOpenAIClient:
    def __init__(self):
        self.chat = MagicMock()
        self.chat.completions = MagicMock()

    def set_response(self, response: ChatCompletion):
        """
        Set the mock to return a specific response.
        :param response: A ChatCompletion response to return.
        """
        self.chat.completions.create.return_value = response

    def set_sequential_responses(self, responses: list[ChatCompletion]):
        """
        Set the mock to return different responses sequentially.
        :param responses: A list of ChatCompletion responses to return in order.
        """
        self.chat.completions.create.side_effect = responses

    def assert_create_called_with(self, **kwargs):
        self.chat.completions.create.assert_called_with(**kwargs)


# Initialize the mock client
client = MockOpenAIClient()

# Set a sequence of mock responses
client.set_sequential_responses(
    [
        create_mock_response(
            {"role": "assistant", "content": "First response"},
            [
                {
                    "name": "process_refund",
                    "args": {"item_id": "item_123", "reason": "too expensive"},
                }
            ],
        ),
        create_mock_response({"role": "assistant", "content": "Second"}),
    ]
)

# This should return the first mock response
first_response = client.chat.completions.create()
print(
    first_response.choices[0].message
)  # Outputs: role='agent' content='First response'

# This should return the second mock response
second_response = client.chat.completions.create()
print(
    second_response.choices[0].message
)  # Outputs: role='agent' content='Second response'



from swarm.util import function_to_json, json_to_function_args
from typing import Any, Callable, Dict, List
import inspect
from pydantic import BaseModel

def test_basic_function():
    def basic_function(arg1, arg2):
        return arg1 + arg2

    result = function_to_json(basic_function)
    assert result == {
        "type": "function",
        "function": {
            "name": "basic_function",
            "description": "",
            "parameters": {
                "type": "object",
                "properties": {
                    "arg1": {"type": "string"},
                    "arg2": {"type": "string"},
                },
                "required": ["arg1", "arg2"],
            },
        },
    }


def test_complex_function():
    def complex_function_with_types_and_descriptions(
        arg1: int, arg2: str, arg3: float = 3.14, arg4: bool = False
    ):
        """This is a complex function with a docstring."""
        pass

    result = function_to_json(complex_function_with_types_and_descriptions)
    assert result == {
        "type": "function",
        "function": {
            "name": "complex_function_with_types_and_descriptions",
            "description": "This is a complex function with a docstring.",
            "parameters": {
                "type": "object",
                "properties": {
                    "arg1": {"type": "integer"},
                    "arg2": {"type": "string"},
                    "arg3": {"type": "number"},
                    "arg4": {"type": "boolean"},
                },
                "required": ["arg1", "arg2"],
            },
        },
    }


def test_pydantic_model_function():
    from pydantic import BaseModel

    class MyModel(BaseModel):
        field1: str
        field2: int

    def function_using_pydantic(model: MyModel):
        return model.field1

    result = function_to_json(function_using_pydantic)
    assert result == {
        "type": "function",
        "function": {
            "name": "function_using_pydantic",
            "description": "",
            "parameters": {
                "type": "object",
                "properties": {
                    "model": {
                        "type": "object",
                        "properties": {
                            "field1": {"type": "string"},
                            "field2": {"type": "integer"},
                        },
                        "required": ["field1", "field2"],
                    },
                },
                "required": ["model"],
            },
        },
    }


def test_pydantic_model_list_function():
    from pydantic import BaseModel
    from typing import List

    class Item(BaseModel):
        name: str
        quantity: int

    def function_with_model_list(items: List[Item]):
        return [item.name for item in items]

    result = function_to_json(function_with_model_list)
    assert result == {
        "type": "function",
        "function": {
            "name": "function_with_model_list",
            "description": "",
            "parameters": {
                "type": "object",
                "properties": {
                    "items": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "name": {"type": "string"},
                                "quantity": {"type": "integer"},
                            },
                            "required": ["name", "quantity"],
                        },
                    },
                },
                "required": ["items"],
            },
        },
    }


def test_json_to_function_args_single_pydantic_model():
    class UserRequest(BaseModel):
        name: str
        age: int

    def sample_func(request: UserRequest):
        pass

    json_args = {
        "request": {
            "name": "Alice",
            "age": 30
        }
    }

    args = json_to_function_args(sample_func, json_args)

    assert isinstance(args['request'], UserRequest)
    assert args['request'].name == "Alice"
    assert args['request'].age == 30


def test_json_to_function_args_list_of_pydantic_models():
    class Item(BaseModel):
        id: int
        description: str

    def sample_func(items: List[Item]):
        pass

    json_args = {
        "items": [
            {"id": 1, "description": "Item 1"},
            {"id": 2, "description": "Item 2"}
        ]
    }

    args = json_to_function_args(sample_func, json_args)

    assert isinstance(args['items'], list)
    assert all(isinstance(item, Item) for item in args['items'])
    assert args['items'][0].id == 1
    assert args['items'][0].description == "Item 1"
    assert args['items'][1].id == 2
    assert args['items'][1].description == "Item 2"


def test_json_to_function_args_mixed_arguments():
    class Address(BaseModel):
        street: str
        city: str

    class Order(BaseModel):
        order_id: int
        items: List[str]

    def sample_func(address: Address, orders: List[Order], confirmed: bool = False):
        pass

    json_args = {
        "address": {
            "street": "123 Main St",
            "city": "Metropolis"
        },
        "orders": [
            {"order_id": 101, "items": ["apple", "banana"]},
            {"order_id": 102, "items": ["orange"]}
        ],
        "confirmed": True
    }

    args = json_to_function_args(sample_func, json_args)

    # Assert Address
    assert isinstance(args['address'], Address)
    assert args['address'].street == "123 Main St"
    assert args['address'].city == "Metropolis"

    # Assert Orders
    assert isinstance(args['orders'], list)
    assert len(args['orders']) == 2
    assert isinstance(args['orders'][0], Order)
    assert args['orders'][0].order_id == 101
    assert args['orders'][0].items == ["apple", "banana"]
    assert isinstance(args['orders'][1], Order)
    assert args['orders'][1].order_id == 102
    assert args['orders'][1].items == ["orange"]

    # Assert Confirmed
    assert args['confirmed'] is True


# Additional Unit Tests for Lists of Primitive Types

def test_function_to_json_with_list_of_strings():
    def function_with_str_list(names: List[str]):
        return ", ".join(names)

    result = function_to_json(function_with_str_list)
    assert result == {
        "type": "function",
        "function": {
            "name": "function_with_str_list",
            "description": "",
            "parameters": {
                "type": "object",
                "properties": {
                    "names": {
                        "type": "array",
                        "items": {"type": "string"},
                    },
                },
                "required": ["names"],
            },
        },
    }


def test_function_to_json_with_list_of_integers():
    def function_with_int_list(numbers: List[int]):
        return sum(numbers)

    result = function_to_json(function_with_int_list)
    assert result == {
        "type": "function",
        "function": {
            "name": "function_with_int_list",
            "description": "",
            "parameters": {
                "type": "object",
                "properties": {
                    "numbers": {
                        "type": "array",
                        "items": {"type": "integer"},
                    },
                },
                "required": ["numbers"],
            },
        },
    }


def test_function_to_json_with_list_of_floats():
    def function_with_float_list(values: List[float]):
        return [v * 2 for v in values]

    result = function_to_json(function_with_float_list)
    assert result == {
        "type": "function",
        "function": {
            "name": "function_with_float_list",
            "description": "",
            "parameters": {
                "type": "object",
                "properties": {
                    "values": {
                        "type": "array",
                        "items": {"type": "number"},
                    },
                },
                "required": ["values"],
            },
        },
    }


def test_json_to_function_args_with_list_of_strings():
    def sample_func(names: List[str]):
        pass

    json_args = {
        "names": ["Alice", "Bob", "Charlie"]
    }

    args = json_to_function_args(sample_func, json_args)

    assert isinstance(args['names'], list)
    assert all(isinstance(name, str) for name in args['names'])
    assert args['names'] == ["Alice", "Bob", "Charlie"]


def test_json_to_function_args_with_list_of_integers():
    def sample_func(numbers: List[int]):
        pass

    json_args = {
        "numbers": [1, 2, 3, 4, 5]
    }

    args = json_to_function_args(sample_func, json_args)

    assert isinstance(args['numbers'], list)
    assert all(isinstance(num, int) for num in args['numbers'])
    assert args['numbers'] == [1, 2, 3, 4, 5]


def test_json_to_function_args_with_list_of_floats():
    def sample_func(values: List[float]):
        pass

    json_args = {
        "values": [1.5, 2.5, 3.75]
    }

    args = json_to_function_args(sample_func, json_args)

    assert isinstance(args['values'], list)
    assert all(isinstance(val, float) for val in args['values'])
    assert args['values'] == [1.5, 2.5, 3.75]



from swarm._sync.types import *

from swarm._async.types import (
  AsyncAgentFunction,
  AsyncAgentInstructions,
  AsyncAgent,
  AsyncResponse,
  AsyncResult,
  AsyncStreamingResponse,
  AsyncMessageStreamingChunk,
  AsyncDelimStreamingChunk,
  AsyncResponseStreamingChunk,
  AsyncDeltaResponseStreamingChunk,
  AsyncStreamingResponse,
)



import re

import qdrant_client
from openai import OpenAI

from swarm import Agent
from swarm.repl import run_demo_loop

# Initialize connections
client = OpenAI()
qdrant = qdrant_client.QdrantClient(host="localhost")

# Set embedding model
EMBEDDING_MODEL = "text-embedding-3-large"

# Set qdrant collection
collection_name = "help_center"


def query_qdrant(query, collection_name, vector_name="article", top_k=5):
    # Creates embedding vector from user query
    embedded_query = (
        client.embeddings.create(
            input=query,
            model=EMBEDDING_MODEL,
        )
        .data[0]
        .embedding
    )

    query_results = qdrant.search(
        collection_name=collection_name,
        query_vector=(vector_name, embedded_query),
        limit=top_k,
    )

    return query_results


def query_docs(query):
    """Query the knowledge base for relevant articles."""
    print(f"Searching knowledge base with query: {query}")
    query_results = query_qdrant(query, collection_name=collection_name)
    output = []

    for i, article in enumerate(query_results):
        title = article.payload["title"]
        text = article.payload["text"]
        url = article.payload["url"]

        output.append((title, text, url))

    if output:
        title, content, _ = output[0]
        response = f"Title: {title}\nContent: {content}"
        truncated_content = re.sub(
            r"\s+", " ", content[:50] + "..." if len(content) > 50 else content
        )
        print("Most relevant article title:", truncated_content)
        return {"response": response}
    else:
        print("No results")
        return {"response": "No results found."}


def send_email(email_address, message):
    """Send an email to the user."""
    response = f"Email sent to: {email_address} with message: {message}"
    return {"response": response}


def submit_ticket(description):
    """Submit a ticket for the user."""
    return {"response": f"Ticket created for {description}"}


def transfer_to_help_center():
    """Transfer the user to the help center agent."""
    return help_center_agent


user_interface_agent = Agent(
    name="User Interface Agent",
    instructions="You are a user interface agent that handles all interactions with the user. Call this agent for general questions and when no other agent is correct for the user query.",
    functions=[transfer_to_help_center],
)

help_center_agent = Agent(
    name="Help Center Agent",
    instructions="You are an OpenAI help center agent who deals with questions about OpenAI products, such as GPT models, DALL-E, Whisper, etc.",
    functions=[query_docs, submit_ticket, send_email],
)

if __name__ == "__main__":
    run_demo_loop(user_interface_agent)
